#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <net/tcp.h>
#include <linux/spinlock.h>
#include <net/checksum.h>

#include <linux/netfilter_ipv4/ip_tables.h>
#include "ipt_IPID.h"

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Jaroslaw Sajko <sloik@man.poznan.pl>");
MODULE_DESCRIPTION("iptables IPID manglong module");

static spinlock_t ipid_lock = SPIN_LOCK_UNLOCKED;

static unsigned int ipid_target(struct sk_buff **pskb,
		const struct net_device *in,
		const struct net_device *out,
		unsigned int hooknum,
		const void *targinfo,
		void *userinfo) {
	

	struct iphdr * iph;
	u_int16_t ipid_diffs[2];
	
	struct ipt_ipid_target_info * ipid_info = (struct ipt_ipid_target_info *)targinfo;
	
	if (!skb_ip_make_writable(pskb, (*pskb)->len))
		return NF_DROP;
	
	iph = (*pskb)->nh.iph;
	
	ipid_diffs[0] = (iph->id)^0xffff;
	if(ipid_info->mode&IPID_MODE_INCREMENTAL) {
		spin_lock_bh(&ipid_lock);
		ipid_diffs[1] = iph->id = htons(ipid_info->master->lastval);
		ipid_info->master->lastval += ipid_info->step;
		spin_unlock_bh(&ipid_lock);
	}
	else if(ipid_info->mode&IPID_MODE_RANDOM) {
		get_random_bytes(&(ipid_diffs[1]), sizeof(u_int16_t));
		ipid_diffs[1] = iph->id = htons(ipid_diffs[1]);
	}
	
	
	iph->check = csum_fold(csum_partial((char *)ipid_diffs, 
				sizeof(ipid_diffs),
				iph->check^0xffff));
#ifdef DEBUG
	if(net_ratelimit()) {
		printk(KERN_INFO "IPID: Id changed %hu -> %hu\n",
				ntohs(ipid_diffs[0]^0xffff), ntohs(iph->id));
	}
#endif

	(*pskb)->nfcache |= NFC_UNKNOWN | NFC_ALTERED;
	return IPT_CONTINUE;
}

static int ipid_checkentry(const char *tablename,
		const struct ipt_entry *e,
		void *targinfo,
		unsigned int targinfosize,
		unsigned int hook_mask) {
	struct ipt_ipid_target_info * ipid_info = (struct ipt_ipid_target_info *)targinfo;
	
	if(strncmp(tablename, "mangle", 6) != 0) {
		printk(KERN_WARNING "IPID: Can only be called from the \"mangle\" table");
		return 0;
	}
	
	if(targinfosize != IPT_ALIGN(sizeof(struct ipt_ipid_target_info))) {
		printk(KERN_WARNING "IPID: targinfosize %u != %Zu\n",
				targinfosize,
				IPT_ALIGN(sizeof(struct ipt_ipid_target_info)));
		return 0;
	}
	ipid_info->master = ipid_info;

	return 1;
}

static void ipid_destroy(void *targinfo, unsigned int targinfosize) {
}

static struct ipt_target ipt_ipid_reg = {
	.name = "IPID",
	.target = ipid_target,
	.checkentry = ipid_checkentry,
	.destroy = ipid_destroy,
	.me = THIS_MODULE,
};

static int __init init(void) {
	return ipt_register_target(&ipt_ipid_reg);
}

static void __exit fini(void) {
	ipt_unregister_target(&ipt_ipid_reg);
}

module_init(init);
module_exit(fini);
